# LUXML = Limited Utility XML (Library)
# by FLACRabbit
# v0.2.1

# Missing major features:
# * Robust error handling
# * Entity support

# DEBUG FUNCTIONS
# ================

LUXMLDEBUG = False

def dprint(t, end='\n'):
	if LUXMLDEBUG:
		print(t, end=end)
	else:
		pass

def crepr(contents):
	# content list repr
	cs = ''
	for c in contents:
		if type(c) == str:
			sc = '"' + c
			sc = sc.replace('\n', '\\n')
			if len(sc) > 15:
				sc = sc[:12] + '...'
			sc += '"'
			cs += sc
		else:
			cs += repr(c)
		cs += ', '
	return cs[:-2]

# OTHER FUNCTIONS
# ===============

def default_relevant(text):
	for c in text:
		if c not in [' ', '\n']:
			return True
	return False

def always_relevant(text):
	return True

def level_split(text, splitters, opener, closer):
	level = 0
	out = []
	current = ''
	for c in text:
		if level > 0:
			if c == closer:
				level -= 1
			elif c == opener:
				level += 1
			current += c
		else:
			if c == opener:
				level += 1
				current += c
			elif c in splitters:
				out.append(current)
				current = ''
			else:
				current += c
	if current: out.append(current)
	return out

# DATA STRUCTURES
# ===============

class LUXMLNode:
	def __init__(self, name='', properties={}, contents=[], header=False):
		self.name = name
		self.properties = properties
		self.contents = contents
		self.header = header
	def __repr__(self):
		return f'{self.name}[{crepr(self.contents)}]'

class LUXMLTag:
	def __init__(self):
		self.close = False
		self.open = False
		self.properties = {}
		self.name = ''
		self.header = False
	def __repr__(self):
		cos = ''
		if self.close:
			cos += 'close'
		if self.open:
			cos += 'open'
		if self.close and self.open:
			cos = 'open+close'
		return f'LUXMLTag({self.name}, {cos}, {self.properties})'

# PARSER
# ======

class LUXMLParser:
	def __init__(self, text, relevant=default_relevant):
		self.text = text
		self.relevant = relevant
	def parse(self, text=None):
		out = []

		if text == None:
			text = self.text
		if text == '':
			return []

		dprint(f"LUXML Parsing {text[:10]}...")

		cont = True

		level = 0
		node_contents = ''

		node_name = ''
		node_properties = {}

		i = 0

		while cont:
			c = text[i]
			if c == '<':
				# consume WHOLE tag
				tag_string = ''
				end = i
				while text[end] != '>':
					end += 1
				tag_string = text[i:end+1]
				i = end
				tag = self.parse_tag(tag_string)
				if tag == None:
					# comment
					continue
				if tag.open and tag.close:
					dprint("Handling open+close tag...")
					if level == 0:
						dprint("Open+close at toplevel.")
						out += [LUXMLNode(tag.name, tag.properties, [])]
						if tag.header:
							out[-1].header = True
					else:
						node_contents += tag_string
				elif tag.open:
					dprint(f"Handling open tag [{level}]->[{level+1}]...")
					if level == 0:
						dprint("Open at toplevel.")
						node_name = tag.name
						node_properties = tag.properties
						if node_contents and self.relevant(node_contents): out += [node_contents]
						node_contents = ''
					else:
						node_contents += tag_string
					level += 1
				elif tag.close:
					dprint(f"Handling close tag [{level}]->[{level-1}]...")
					level -= 1
					if level == 0:
						dprint("Close at toplevel.")
						out += [LUXMLNode(node_name, node_properties, self.parse(node_contents))]
						node_contents = ''
					else:
						node_contents += tag_string
			elif c != '<' or level >= 1:
				# add if either we're on the top level handling some text (c != '<') or we're inside a node (level >= 1)
				node_contents += c
			i += 1
			if i >= len(text):
				cont = False

		if level != 0:
			out += [LUXMLNode(node_name, node_properties, self.parse(node_contents))]
			node_contents = ''

		if node_contents and self.relevant(node_contents): out += [node_contents] # in case we ended with text

		return out
	def parse_tag(self, tag_string):
		dprint(f"Tag parsing from {tag_string}...")
		tag = LUXMLTag()
		if tag_string[1] == '?':
			tag_string = tag_string[0] + tag_string[2:]
			tag_string = tag_string[:-2] + tag_string[-1]
			tag.header = True
			tag.open = True
			tag.close = True
		if tag_string[1] == '!':
			# comment
			return None
		if tag_string[1] == '/' or (len(tag_string) > 2 and tag_string[-2] == '/'):
			tag.close = True
		if tag_string[1] != '/':
			tag.open = True
		if tag_string[1] == '/':
			tag_string = tag_string[0] + tag_string[2:]
		if tag_string[-2] == '/':
			tag_string = tag_string[:-2] + tag_string[-1]
		for e, item in enumerate(level_split(tag_string[1:-1], [' ', '\t', '\n'], '"', '"')):
			if not item:
				continue
			if e == 0:
				tag.name = item
			else:
				dprint(f"Item = {item}...")
				tag.properties[item.split('=')[0]] = item.split('=')[1][1:-1]
		dprint(f"\tParsed as {tag}.")
		return tag

def luxml_parse(text):
	p = LUXMLParser(text)
	return p.parse()

def luxml_parse_file(file):
	fh = open(file)
	content = fh.read()
	fh.close()
	p = LUXMLParser(content)
	return p.parse()

# WRITER
# ======

def write_node(node):
	if type(node) == str:
		return node
	out = ''
	op = '<'
	if node.header:
		op = '<?'
	out += f'{op}{node.name}'
	for n in node.properties.keys():
		out += f' {n}="{node.properties[n]}"'
	if node.contents == []:
		if node.header:
			out += '?>'
		else:
			out += '/>'
	else:
		out += '>'
		for n in node.contents:
			if type(n) == str:
				out += n
			else:
				out += write_node(n)
		out += f'</{node.name}>\n'
	return out

class LUXMLWriter:
	def __init__(self, nodes):
		self.nodes = nodes
	def write(self):
		return '\n'.join([write_node(x) for x in self.nodes])

def luxml_generate_prolog(version=1.0, encoding='UTF-8'):
	return LUXMLNode('xml', {'version':str(version), 'encoding':str(encoding)}, [], header=True)

def luxml_write(nodes):
	l = LUXMLWriter(node)
	return l.write()

def luxml_write_file(nodes, filename):
	l = LUXMLWriter(nodes)
	content = l.write()
	fh = open(filename, 'w')
	fh.write(content)
	fh.close()

# NODE TOOLS
# ==========

def luxml_node_subnodes(node):
	out = []
	for c in node.contents:
		if type(c) == LUXMLNode:
			out.append(c)
	return out

def luxml_node_subnodes_matching(node, name):
	out = []
	for c in node.contents:
		if type(c) == LUXMLNode and c.name == name:
			out.append(c)
	return out

def luxml_node_recursive_subnodes(node, include_top=False):
	out = []
	if include_top: out += [node]
	for c in node.contents:
		if type(c) == LUXMLNode:
			out.extend(luxml_node_recursive_subnodes(c, True))
	return out

def luxml_node_recursive_subnodes_matching(node, name, include_top=False):
	out = []
	if include_top and node.name == name: out += [node]
	for c in node.contents:
		if type(c) == LUXMLNode and c.name == name:
			out.extend(luxml_node_recursive_subnodes_matching(c, name, True))
	return out

def luxml_node_texts(node):
	out = []
	for c in node.contents:
		if type(c) == str:
			out.append(c)
	return out

def luxml_node_all_text(node, joiner=''):
	"""Merges all the toplevel text together."""
	return joiner.join(luxml_node_texts(node))

def luxml_node_recursive_text(node):
	"""Merges all the text from this node and every subnode together."""
	out = ''
	for c in node.contents:
		if type(c) == str:
			out += c
		elif type(c) == LUXMLNode:
			out += luxml_node_recursive_text(c)
	return out
